from os import linesep
from json import JSONEncoder
from csv import DictWriter as CsvDictWriter


def formatter_factory(fp, format):
    formatter = CsvFormatter(fp)

    if 'txt' in format:
        formatter = TextFormatter(fp)
    elif 'json' in format:
        formatter = JsonFormatter(fp)
    return formatter


class Formatter():
    def __init__(self, fp):
        self.fp = fp

    def iterout(self, iterable):
        if iterable is None:
            return
        try:
            header = next(iterable)

            if hasattr(header, 'keys'):
                self.begin(header.keys())
            else:
                self.begin(header)
            self.out(header)
            for item in iterable:
                self.out(item)
            self.end()
        except StopIteration:
            pass

    def out(self, d):
        raise NotImplementedError('Formatters must provide a row method.')

    def begin(self, fields):
        return self.out(fields)

    def end(self):
        pass


class TextFormatter(Formatter):
    def begin(self, fields):
        pass

    def out(self, d):
        self.fp.write(str(d) + linesep)


class JsonFormatter(Formatter):
    def begin(self, fields):
        self.encoder = JSONEncoder(indent=2, separators=(', ', ': '))

    def iterout(self, iterable):
        self.begin([])
        for chunk in self.encoder.iterencode(list(iterable)):
            self.fp.write(chunk)

    def out(self, d):
        self.fp.write(self.encoder.encode(d))


class CsvFormatter(Formatter):
    def begin(self, fields):
        self.writer = CsvDictWriter(self.fp, fieldnames=fields)
        self.writer.writeheader()

    def out(self, d):
        self.writer.writerow(d)



from ..patterns import PatternList
from ..providers import ProviderList, DefaultProviders
from ..send import Sender


class InfoCommand(object):
    def __call__(self, ns):
        return ns.formatter.iterout(self.execute(ns))

    def execute(self, ns):
        providers = ProviderList() + DefaultProviders
        patterns = PatternList(ns.commands[:-1])
        sender = None

        if len(ns.commands[-1:]):
            sender = Sender(ns.commands[-1:].pop())

        if ns.providers:
            providers += ProviderList.from_str(ns.providers)
        if ns.patterns:
            patterns += PatternList.from_path(ns.patterns)

        for n in ns.__dict__.items():
            yield ['arg', n[0], n[1]]
        for provider in providers:
            yield ['provider', provider]
        for pattern in patterns:
            yield ['pattern', pattern]
        yield ['sender', sender]



from ..gen import Generator
from ..patterns import PatternList
from ..providers import ProviderList, DefaultProviders
from ..send import Sender


class SendCommand(object):
    def __call__(self, ns):
        return ns.formatter.iterout(self.execute(ns))

    def execute(self, ns):
        assert 0 != len(ns.commands), \
            'Usage is: gensend send "pattern1" "pattern2" "...pattern%{INT}" ' \
            '"scheme://netloc/path;parameters?query#fragment"'

        providers = ProviderList() + DefaultProviders
        generator = Generator()
        patterns = PatternList(ns.commands[:-1])
        sender = Sender(ns.commands[-1:].pop())

        if ns.providers:
            providers += ProviderList.from_str(ns.providers)
        if ns.patterns:
            patterns += PatternList.from_path(ns.patterns)
        if not ns.noseed:
            generator.seed_providers(ns.seed)
        generator.add_providers(providers)

        with sender.session() as s:
            for generatable in generator.generate(patterns, rounds=ns.number):
                data = generatable()
                s.send(data)
                yield data



import sys
import os
import argparse
from .info import InfoCommand
from .gen import GenCommand
from .send import SendCommand
from ..formatters import formatter_factory


EXIT_CODE_BAD_ARG = 1
EXIT_CODE_GENERAL_ERR = 2


class ArgumentParserError(Exception):
    pass


class ThrowingArgumentParser(argparse.ArgumentParser):
    def error(self, message):
        raise ArgumentParserError(message)


class Cmd(object):

    def __init__(self):
        self.argparser = None

    def run(self):
        self.build_argparser()
        self.build_commands()
        self.run_argparser()

    def build_argparser(self):

        # Entry parser
        self.argparser = ThrowingArgumentParser()
        self.argparser.add_argument(
            '--load-providers', '-l', default=None, dest='providers',
            help='an optional file path or importable module name that may be'
                 ' included and added to the list of default providers.')
        self.argparser.add_argument(
            '--patterns', '-p', default=None, dest='patterns',
            help='optional file containing pattern strings separated by new'
                 ' lines when .txt, or a json array of patterns when .json')
        self.argparser.add_argument(
            '--format', '-f', default='txt', dest='format',
            help='format for output data, one of txt, json or csv default: txt')
        self.argparser.add_argument(
            '--debug', '-d', default=False, action='store_true', dest='debug',
            help='debug flag for additional troubleshooting information')
        self.argparser.add_argument(
            '--number', default=1, dest='number', type=int,
            help='number of times to generate pattern, default: 1')
        self.argparser.add_argument(
            '--persistent', default=True, dest='persistent', action='store_true',
            help='when sending multiple message keep the connection open. note when using'
                 ' this option you should terminate your pattern if needed. default: true')
        gen_parser_group = self.argparser.add_mutually_exclusive_group()
        gen_parser_group.add_argument(
            '--noseed', default=False, dest='noseed', action='store_true',
            help='do not seed the pseudo-random number generator to produce non-deterministic data')
        gen_parser_group.add_argument(
            '--seed', default=1, dest='seed', type=int,
            help='seed the pseudo-random number generator, default: 1')

    def build_commands(self):

        # Sub parser for commands
        sub_parsers = self.argparser.add_subparsers(title='commands')

        # Info command
        info_parser = sub_parsers.add_parser(
            'info',
            description='dump information to stdout and exit')
        info_parser.set_defaults(func=self.command_info)
        info_parser.add_argument('commands', nargs='*')

        # Gen command
        gen_parser = sub_parsers.add_parser(
            'gen',
            description='generate data',
            help='ex: gensend gen pattern1 ... patternN')
        gen_parser.set_defaults(func=self.command_gen)
        gen_parser.add_argument('commands', nargs='*')

        # Send command
        send_parser = sub_parsers.add_parser(
            'send',
            description='send data',
            help='ex: gensend send pattern1 ... patternN scheme://netloc/path;parameters?query#fragment')
        send_parser.set_defaults(func=self.command_send)
        send_parser.add_argument('commands', nargs='*')

    def run_argparser(self):
        ns = None

        try:
            ns = self.argparser.parse_args()
            self.command(ns)
        except ArgumentParserError, e:
            self.oops(str(e))
        return ns

    def oops(self, message):
        print '\nOops: ' + message
        print 'Maybe you could try: gensend -h'
        print "\n----\n"
        self.argparser.print_help()
        self.error('', EXIT_CODE_GENERAL_ERR)

    def error(self, msg, code):
        sys.stderr.write(msg + "\n")
        sys.stderr.flush()
        os._exit(code)

    def command(self, ns):
        ns.formatter = formatter_factory(sys.stdout, ns.format)
        ns.func(ns)

    def command_info(self, ns):
        """ Command: gensend info"""
        infocmd = InfoCommand()
        infocmd(ns)

    def command_gen(self, ns):
        """ Command: gensend gen"""
        gencmd = GenCommand()
        gencmd(ns)

    def command_send(self, ns):
        """ Command: gensend send"""
        sendcmd = SendCommand()
        sendcmd(ns)



import os
import json


class PatternList(list):

    @classmethod
    def from_path(cls, pattern_path):
        pattern_path = os.path.abspath(pattern_path)
        patterns = []

        with open(pattern_path) as f:
            if pattern_path.endswith('.txt'):
                patterns = [pattern.strip() for pattern in f]
            elif pattern_path.endswith('.json'):
                patterns = json.load(f)
        assert len(patterns), 'Unable to parse patterns file at {0}.'.format(pattern_path)
        return PatternList(patterns)



from urlparse import urlparse
import socket


class Sender(object):

    @classmethod
    def from_components(cls, components):
        raise cls(**components)

    def __init__(self, sendable):
        self.sendable = sendable
        self.transport = Transport.from_sendable(sendable)

    def __enter__(self):
        self.transport.open()
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.transport.close()

    def session(self):
        return SenderSession(self)

    def send(self, data):
        self.transport.send(data)


class SenderSession(object):
    def __init__(self, sender):
        self.sender = sender

    def __enter__(self):
        return self.sender.__enter__()

    def __exit__(self, *args):
        return self.sender.__exit__(*args)


class Transport(object):
    TRANSPORTS = []

    @staticmethod
    def register(transport):
        Transport.TRANSPORTS.append(transport)

    @staticmethod
    def from_sendable(sendable):
        components = urlparse(sendable)
        transports = [t for t in Transport.TRANSPORTS if t.SCHEME == components.scheme]

        if not len(transports):
            raise Exception('Expected at least one matching transport')
        return transports.pop(0).from_components(components)

    @classmethod
    def from_components(cls, components):
        props = ['username', 'password', 'hostname', 'port']
        components_dict = dict([(k, getattr(components, k)) for k in props], **components._asdict())

        return cls(**components_dict)

    def __enter__(self):
        self.connect()
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.disconnect()

    def open(self):
        raise NotImplementedError('Transport missing open method')

    def close(self):
        raise NotImplementedError('Transport missing close method')

    def send(self, data):
        raise NotImplementedError('Transport missing send method')


class SocketTransport(Transport):

    @classmethod
    def __init__(cls, hostname=None, port=None, **kwargs):
        cls.address = (hostname, port)
        cls.socket = socket.socket(socket.AF_INET, cls.SOCKET_TYPE)

    def open(self):
        self.socket.connect(self.address)

    def close(self):
        self.socket.close()

    def send(self, data):
        sent = self.socket.send(data)
        if sent == 0:
            raise RuntimeError('unable to send to socket')


@Transport.register
class UdpTransport(SocketTransport):
    SCHEME = 'udp'
    SOCKET_TYPE = socket.SOCK_DGRAM


@Transport.register
class TcpTransport(SocketTransport):
    SCHEME = 'tcp'
    SOCKET_TYPE = socket.SOCK_STREAM



from ..gen import Generator
from ..patterns import PatternList
from ..providers import ProviderList, DefaultProviders


class GenCommand(object):
    def __call__(self, ns):
        return ns.formatter.iterout(self.execute(ns))

    def execute(self, ns):
        providers = ProviderList() + DefaultProviders
        generator = Generator()
        patterns = PatternList(ns.commands[:])

        if ns.providers:
            providers += ProviderList.from_str(ns.providers)
        if ns.patterns:
            patterns += PatternList.from_path(ns.patterns)
        if not ns.noseed:
            generator.seed_providers(ns.seed)

        assert len(patterns) and len(providers), \
            'Usage is: gensend gen "pattern1" "pattern2" "...pattern%{INT}"'

        generator.add_providers(providers)

        for generatable in generator.generate(patterns, rounds=ns.number):
            yield generatable()



from faker import Faker
from .provider import Provider, ProviderList, ProviderAccessor  # flake8: noqa
from .common import Common, Control, WrappedPackages, WrappedFuncs


DefaultProviders = ProviderList([Common(), Control(), WrappedPackages(), WrappedFuncs(), Faker()])



import unittest
from ..gen import Generator
from ..providers import DefaultProviders


class TestingHelper(unittest.TestCase):

    def __init__(self, *args, **kwargs):
        super(TestingHelper, self).__init__(*args, **kwargs)
        self.generator = Generator(provider_list=DefaultProviders)
        self.generator.seed_providers(1)

    def assert_patterns(self, patterns):
        for (pattern, expected) in patterns:
            self.assert_pattern(pattern, expected)

    def assert_pattern(self, pattern, expected):
        result = self.generator.run(pattern)
        assert expected == result



from . import TestingHelper


class TestTokenizer(TestingHelper):

    def test_literal_tokens(self):
        patterns = [
            ['literal ":"', 'literal ":"'],
            ['literal "\:"', 'literal ":"'],
            ['literal ","', 'literal ","'],
            ['literal "\,"', 'literal ","'],
            ['literal "}"', 'literal "}"'],
            ['literal "\}"', 'literal "}"'],
            ['literal "\\\\"', 'literal "\\"'],
            ['literal "%"', 'literal "%"'],
            ['literal "{"', 'literal "{"'],
            ['literal "\\%{"', 'literal "%{"'],
        ]
        self.assert_patterns(patterns)

    def test_escaping_edge(self):
        patterns = [
            ['is backslash "\\"', 'is backslash "\\"'],
            ['is backslash "\\\\}" followed by close tok', 'is backslash "\\}" followed by close tok'],
            ['is backslash "\\\\%{int}" followed by valid open tok',
                'is backslash "\\4865782901354085937" followed by valid open tok'],
            ['slash city "\\\\\\%{int}"', 'slash city "\\%{int}"'],
        ]
        self.assert_patterns(patterns)


class TestParser(TestingHelper):

    def test_nesting(self):
        patterns = [
            ['depth 0 = %{FIRST:1,2,3}', 'depth 0 = 1'],
            ['depth 1 = %{FIRST:%{FIRST:4,5,6},1,2,3}', 'depth 1 = 4'],
            ['depth 2 = %{FIRST:%{FIRST:%{FIRST:%{FIRST:7,8,9},2,3},4,5,6},1,2,3}', 'depth 2 = 7'],
            ['depth 3 = %{FIRST:%{FIRST:%{FIRST:%{FIRST:%{FIRST:10,11,12},8,9},2,3},4,5,6},1,2,3}', 'depth 3 = 10'],
        ]
        self.assert_patterns(patterns)



import sys
import logging


class GensendLogger(logging.Logger):
    def __call__(self, *args, **kwargs):
        return self.debug(*args, **kwargs)


def configure_logger(name, level=logging.DEBUG):
    logging.captureWarnings(True)
    logging.setLoggerClass(GensendLogger)

    handler = logging.StreamHandler(sys.stdout)
    handler.setLevel(level)
    handler.setFormatter(logging.Formatter('[%(asctime)s] %(name)s.%(levelname)s: %(message)s'))

    logger = logging.getLogger(name)
    logger.setLevel(level)
    logger.addHandler(handler)

    return logger


log = configure_logger('gensend', logging.DEBUG)



from .providers import ProviderList, ProviderAccessor


class Generator(ProviderAccessor):
    def __init__(self, provider_list=None):
        self.tokenizer = Tokenizer()
        self.parser = Parser()
        self.provider_list = ProviderList()
        self.patterns = []

        if provider_list:
            self.add_providers(provider_list)

    def compile(self, pattern, provider_list=None):
        if provider_list is None:
            provider_list = self.provider_list

        generatable = Generatable(self.provider_list, pattern)
        tokens = self.tokenizer.tokenize(pattern)
        ctx = ParserCtx(generatable, tokens)
        self.parser.parse(ctx)

        return generatable

    def run(self, pattern):
        return self.compile(pattern)()

    def generate(self, patterns, rounds=1):
        for i in range(rounds):
            for pattern in patterns:
                yield self.compile(pattern)


class Printable(object):
    def __str__(self):
        space = "  "
        height = self.height + 1
        label = self.__class__.__name__
        spacer = space * height
        out = spacer

        if 0 == height:
            label += "(Root)"
        else:
            label += "(" + str(height) + ")"
        out += label

        if len(self.payload):
            out += "\n" + spacer + (space * 2) + "-> " + str(self.payload)

        out += "\n"
        for child in self.children:
            out += str(child)
        return out

    def __repr__(self):
        return str(self.__class__.__name__) + "(" + str(self.__dict__) \
            .replace("'", '').replace("{", '').replace("}", ', ').rstrip(', ') + ")"


class Token(Printable):
    TOK_STR = 'TOK_STR'  # '*'
    TOK_OPEN = 'TOK_OPEN'  # '%{'
    TOK_CLOSE = 'TOK_CLOSE'  # '}'
    TOK_SEP = 'TOK_SEP'  # ':'
    TOK_ARG_SEP = 'TOK_ARG_SEP'  # ','

    def __init__(self, tok, val, pos):
        self.tok = tok
        self.val = val
        self.pos = pos


class Tokenizer(Printable):
    def __init__(self):
        self.buf = ''
        self.end = 0
        self.pos = 0

    def at(self, pos):
        if (pos >= self.end) or pos < 0:
            return None
        return self.buf[pos]

    @property
    def lo(self):
        return self.at(self.pos)

    @property
    def lb(self):
        return self.at(self.pos-1)

    @property
    def la(self):
        return self.at(self.pos+1)

    def scan(self, pattern):
        self.pos = 0
        self.buf = pattern
        self.end = len(pattern)

        while 1:
            token = self.tok()
            if not token:
                break
            yield token

    def tok(self):
        token = None

        if '\\' == self.lo and self.la in ('%', ',', ':', '{', '}', '\\'):
            self.pos += 1
            token = Token(Token.TOK_STR, self.lo, self.pos)
        elif '%' == self.lo and '{' == self.la:
            self.pos += 1
            token = Token(Token.TOK_OPEN, None, self.pos)
        elif '}' == self.lo:
            token = Token(Token.TOK_CLOSE, None, self.pos)
        elif ':' == self.lo:
            token = Token(Token.TOK_SEP, None, self.pos)
        elif ',' == self.lo:
            token = Token(Token.TOK_ARG_SEP, None, self.pos)
        elif self.lo:
            token = Token(Token.TOK_STR, self.lo, self.pos)
        self.pos += 1
        return token

    def tokenize(self, pattern):
        return list(self.scan(pattern))


class Node(Printable):
    def __call__(self):
        return str(self.execute())

    def __init__(self, payload=None):
        if payload is None:
            payload = ''
        self.payload = payload
        self.parent = None
        self.children = []

    @property
    def top(self):
        parent = self

        while parent.parent:
            parent = parent.parent
        return parent

    @property
    def height(self):
        height = -1
        parent = self

        while parent:
            height += 1
            parent = parent.parent
        return height

    def add(self, node):
        node.parent = self
        self.children.append(node)

    def execute(self):
        return self.payload


class StringNode(Node):
    def execute(self):
        result = self.payload

        for child in self.children:
            result += child()
        return result


class ExprNode(Node):
    def execute(self):
        if not len(self.payload):
            return ''
        expr = self.payload
        expr_args = []

        for child in self.children:
            expr_args.append(child())
        return self.top.call_provider(expr, expr_args)


class Generatable(StringNode, ProviderAccessor, Printable):
    def __init__(self, provider_list, source):
        Node.__init__(self)
        self.provider_list = provider_list
        self.source = source


class Parser(Printable):
    def __init__(self):
        self.token_call_map = {
            Token.TOK_OPEN: 'do_tok_open',
            Token.TOK_SEP: 'do_tok_sep',
            Token.TOK_ARG_SEP: 'do_tok_arg_sep',
            Token.TOK_CLOSE: 'do_tok_close',
            Token.TOK_STR: 'do_tok_str',
        }

    def parse(self, ctx):
        self.begin(ctx)
        self.walk(ctx)
        self.end(ctx)

    def begin(self, ctx):
        ctx.begin()

    def walk(self, ctx):
        for token in ctx.tokens:
            if token.val:
                getattr(ctx, self.token_call_map[token.tok])(token.val)
            else:
                getattr(ctx, self.token_call_map[token.tok])()

    def end(self, ctx):
        ctx.end()


class ParserCtx(Printable):
    STATE_EXIT = 'STATE_EXIT'
    STATE_IDLE = 'STATE_IDLE'
    STATE_STR = 'STATE_STR'
    STATE_EXPR = 'STATE_EXPR'
    STATE_EXPR_ARG = 'STATE_EXPR_ARG'

    def __init__(self, generatable, tokens):
        self.generatable = generatable
        self.tokens = tokens
        self.stack = [generatable]
        self.state = [ParserCtx.STATE_IDLE]

    @property
    def stack_top(self):
        return self.stack[-1] if len(self.stack) else None

    @property
    def state_top(self):
        return self.state[-1] if len(self.state) else ParserCtx.STATE_EXIT

    def begin(self):
        self.enter(ParserCtx.STATE_STR)

    def end(self):
        while 1:
            popped = self.stack.pop()
            if not self.stack_top:
                break
            self.stack_top.add(popped)

    def enter(self, to_state):
        old_state = self.state_top
        cur_state = to_state
        cur_node = None

        assert ParserCtx.STATE_IDLE != to_state, 'unable to transition to idle'
        assert cur_state != old_state, 'unable to transition to same state'

        if ParserCtx.STATE_STR == old_state:
            self.exit()
        if ParserCtx.STATE_STR == cur_state:
            cur_node = StringNode()
        elif ParserCtx.STATE_EXPR == cur_state:
            cur_node = ExprNode()
        elif ParserCtx.STATE_EXPR_ARG == cur_state:
            cur_node = StringNode()

        assert not (cur_node is None), 'unable to derive a new node from state transition'

        self.state.append(to_state)
        self.stack.append(cur_node)

    def exit(self):
        self.state.pop()
        old_node = self.stack.pop()
        self.stack_top.add(old_node)

    def do_tok_str(self, s):
        if ParserCtx.STATE_IDLE == self.state_top:
            self.enter(ParserCtx.STATE_STR)
        if len(self.stack_top.children):
            self.stack_top.children.append(StringNode(s))
        else:
            self.stack_top.payload += s

    def do_tok_open(self):
        self.enter(ParserCtx.STATE_EXPR)

    def do_tok_sep(self):
        if ParserCtx.STATE_EXPR == self.state_top:
            self.enter(ParserCtx.STATE_EXPR_ARG)
        else:
            self.do_tok_str(':')

    def do_tok_arg_sep(self):
        if ParserCtx.STATE_EXPR_ARG == self.state_top:
            self.exit()
            self.enter(ParserCtx.STATE_EXPR_ARG)
        else:
            self.do_tok_str(',')

    def do_tok_close(self):
        if self.state_top in (ParserCtx.STATE_STR, ParserCtx.STATE_IDLE):
            self.do_tok_str('}')
        else:
            if ParserCtx.STATE_EXPR_ARG == self.state_top:
                self.exit()
            self.exit()



from setuptools import setup, find_packages


setup(
    name='gensend',
    version='0.0.1',
    packages=find_packages(),
    entry_points={'console_scripts': [
        'gensend = gensend.cmd:execute_gensend',
    ]},
    package_data={'': ['*Makefile*']},
    include_package_data=True,
    install_requires=[
      'fake-factory>=0.5.2',
      'netaddr>=0.7.17',
    ],
    url='https://github.com/cstockton/gensend'
)



import sys
import math
import operator
import string
from netaddr import IPNetwork, IPRange, IPGlob
from . import Provider


class WrappedPackages(Provider):
    WRAPPED = {
        operator: {
            None: ['eq', 'ne', 'gt', 'lt', 'ge', 'le'],
            str: ['concat'],
            int: ['add', 'sub', 'div', 'mul', 'mod'],
        },
        math: {
            float: ['ceil', 'fabs', 'factorial', 'floor', 'exp', 'expm1',
                    'log', 'pow', 'sqrt', 'degrees', 'radians'],
        },
        string: {
            str: ['lower', 'upper', 'capitalize', 'swapcase', 'translate',
                  'strip', 'lstrip', 'rstrip']
        }
    }

    def __init__(self):
        Provider.__init__(self)
        for src_object, src_targets in WrappedPackages.WRAPPED.items():
            for target_factories, target_names in src_targets.items():
                for target_name in target_names:
                    setattr(self, target_name, self._bind_func(
                        src_object, target_name, target_factories))


class WrappedFuncs(Provider):
    WRAPPED = {
        min: float,
        max: float,
    }

    def __init__(self):
        Provider.__init__(self)
        for src_func, src_factory in WrappedFuncs.WRAPPED.items():
            setattr(self, src_func.__name__, self._wrap_func(src_func, src_factory))


class Control(Provider, dict):
    def __init__(self):
        dict.__init__(self)
        Provider.__init__(self)
        self['if'] = self._dict_if
        self['repeat'] = self._dict_repeat

    def _dict_if(self, *args):
        """Returns the arguments in the list joined by STR.
            IF:CONDITION,DO,ELSE

        %{IF:False,'IS TRUE'} -> ''
        %{IF:True,'IS TRUE', 'NOT TRUE'} -> 'NOT TRUE'
        """
        call_args = list(args)
        do_cond = call_args.pop(0)
        do_if = call_args.pop(0)
        do_else = ''

        if len(call_args):
            do_else = call_args.pop()
        return do_if if self.truth(do_cond) else do_else

    def _dict_repeat(self, *args):
        """Do a thing N times
            REPEAT:N,DO

        %{REPEAT:10,%{INT}} -> '123 123 ... 123'
        """
        call_args = list(args)
        do_count = call_args.pop(0)
        do_thing = call_args.pop(0)
        do_out = []

        for i in range(int(do_count)):
            do_out.append(do_thing())
        return ''.join(do_out)

    def truth(self, *args):
        def to_truth(val):
            if isinstance(val, (list, tuple)):
                return len(val)
            return self.istrue(val)
        return all(self._arg_factory(to_truth, args))

    def pytruth(self, *args):
        """Performs a python truth value test. bool(arg), if multiple args are
        provided it will truth test them all.
            - https://docs.python.org/2/library/stdtypes.html#truth-value-testing
            PYTRUTH:1

        %{PYTRUTH:0} -> 'False'
        """
        return all(self._arg_factory(bool, args))

    def istrue(self, *args):
        """Strict test for 'true' value test. If multiple args are provided it will
        test them all.
            ISTRUE:true

        %{ISTRUE:true} -> 'True'
        """
        def is_true(val):
            if val is True:
                return True
            val = str(val).lower().strip()
            return val in ('true', 'yes', '1')
        return all(self._arg_factory(is_true, args))

    def isfalse(self, *args):
        def is_true(val):
            if val is False:
                return False
            val = str(val).lower().strip()
            return val in ('false', 'no', '0')
        return all(self._arg_factory(is_true, args))


class Common(Provider):
    """@TODO right now using an untrusted template could possible cause security
    issues the way that it's executed and not boxed in to a specific namespace"""

    def cr(self):
        """Prints carriage return
            \r

        %{LF} -> '\n'
        """
        return "\r"

    def lf(self):
        """Prints line feed
            \n

        %{LF} -> '\n'
        """
        return "\n"

    def crlf(self):
        """Prints carriage return and line feed
            NUL

        %{CRLF} -> '\r\n'
        """
        return "\r\n"

    def nul(self):
        """Prints a null byte
            NUL

        %{NUL} -> 'A'
        """
        return '\x00'

    def types(self, *args):
        """Used for debugging, returns type of each arg.
            TYPES,ARG_1,...,ARG_N

        %{TYPES:A,...,10} -> 'str(A) str(B) ... int(10)'
        """
        return ', '.join(['{0}({1})'.format(type(arg).__name__, arg) for arg in args])

    def join(self, *args):
        """Returns the arguments in the list joined by STR.
            FIRST,JOIN_BY,ARG_1,...,ARG_N

        %{JOIN: ,A,...,F} -> 'A B C ... F'
        """
        call_args = list(args)
        joiner = call_args.pop(0)
        self.random.shuffle(call_args)
        return joiner.join(call_args)

    def first(self, *args):
        """Returns the first value in the list.
            FIRST,ARG_1,...,ARG_N

        %{FIRST:A,...,F} -> 'A'
        """
        return args[0]

    def last(self, *args):
        """Returns the first value in the list.
            FIRST,ARG_1,...,ARG_N

        %{FIRST:A,...,F} -> 'F'
        """
        return args[-1]

    def shuffle(self, *args):
        """Shuffles all arguments and returns them.
            ARG_1,...,ARG_N

        %{SHUFFLE:A, B ,...,F} -> 'CDA B FE'
        """
        call_args = list(args)
        self.random.shuffle(call_args)
        return ''.join(call_args)

    def choice(self, *args):
        """Returns a random item from the arg list
            SEPARATOR,ARG_1,...,ARG_N

        %{CHOICE:A,...,F} -> 'B'
        """
        return self.random.choice(args)

    def float(self, *args):
        """Returns a random float between 0 and 1
            FLOAT

        %{FLOAT} -> '.1234567890'
        """
        return self.random.random()

    def ipglob(self, *args):
        """Returns a random address from within the given ip global
        https://pythonhosted.org/netaddr/api.html#ip-glob-ranges
            IPGLOB:GLOB

        %{IPGLOB:*.*.*.*} -> ''
        """
        call_args = list(args)
        return self.random.choice(IPGlob(call_args.pop(0)))

    def iprange(self, *args):
        """Returns a random address from within the given range of two addresses
            IPRANGE:start,end

        %{IPRANGE:10.0.0.0/8,} -> ''
        """
        call_args = list(args)
        return self.random.choice(IPRange(call_args.pop(0), call_args.pop(0)))

    def ipcidr(self, *args):
        """Returns a random address from within the given cidr notation
            IPCIDR:cidr

        %{IPCIDR:10.0.0.0/8} -> ''
        """
        call_args = list(args)
        return self.random.choice(IPNetwork(call_args.pop(0)))

    def int(self, *args):
        """Returns a random int between -sys.maxint and sys.maxint
            INT

        %{INT} -> '1245123'
        %{INT:10} -> '10000000'
        %{INT:10,20} -> '19'
        """
        return self.random.randint(*self._arg_defaults(args, [-sys.maxint, sys.maxint], int))



import os
from random import Random


class Provider(object):
    def __init__(self):
        self.random = Random()

    def seed(self, seed):
        self.random.seed(seed)

    def _arg_defaults(self, in_args, def_args, factory=None):
        return self._arg_factory(
            factory, [in_args[i] if len(in_args) > i else v for i, v in enumerate(def_args)])

    def _arg_factory(self, factory, args):
        return map(factory, args) if factory else args

    def _wrap_func(self, wrapped, arg_factory=None):
        def wrapper(*args):
            try:
                return wrapped(*self._arg_factory(arg_factory, args))
            except Exception, err:
                return '(FAIL: wrapped func "' + str(wrapped) + '" said ' + str(err) + ')'
        return wrapper

    def _bind_func(self, src_object, attr_name, arg_factory=None):
        return self._wrap_func(getattr(src_object, attr_name), arg_factory)


class ProviderAccessor(object):
    def __init__(self, provider_list=None):
        self.provider_list = ProviderList()

    def call_provider(self, ident, args):
        return self.provider_list.call_provider(ident, args)

    def make_provider_func(self, ident, args):
        return self.provider_list.make_provider_func(ident, args)

    def add_provider(self, provider):
        self.provider_list.append(provider)

    def seed_providers(self, seed):
        self.provider_list.seed_providers(seed)

    def add_providers(self, provider_list):
        for provider in provider_list:
            self.add_provider(provider)


class ProviderList(list):

    @classmethod
    def from_str(cls, provider_str):
        if provider_str.endswith('.py'):
            return cls.from_path(provider_str)
        else:
            return cls.from_module(provider_str)

    @classmethod
    def from_path(cls, provider_str):
        locals_dict = {}
        execfile(os.path.abspath(provider_str), {}, locals_dict)

        return cls.from_mixed_type_list(
            [local_value for local_name, local_value in locals_dict.items()])

    @classmethod
    def from_module(cls, provider_module):
        imported = __import__(provider_module, {}, {})

        return cls.from_mixed_type_list(
            [getattr(imported, name) for name in dir(imported)])

    @classmethod
    def from_mixed_type_list(cls, mixed_type_list):
        providers = ProviderList()

        for mixed_type in mixed_type_list:
            if isinstance(mixed_type, ProviderList):
                providers += mixed_type
        return providers

    def get_provider(self, ident):
        if ident.startswith('_'):
            return lambda *args: \
                '(FAIL: identifer "{0}" begins with an underscore, args were {1})'.format(
                    ident, str(args))
        ident = ident.lower()

        for provider in self:
            if hasattr(provider, ident):
                return getattr(provider, ident)
            if isinstance(provider, dict) and ident in provider:
                return provider[ident]
        return lambda *args: \
            '(FAIL: identifier "{0}" not found, args were {1})'.format(ident, str(args))

    def get_provider_func(self, ident, *args):
        return lambda: str(self.get_provider(ident)(*args))

    def call_provider(self, ident, args):
        return self.get_provider(ident)(*args)

    def seed_providers(self, seed):
        for provider in self:
            provider.seed(seed)
